#include "Fuzzy_Tank.h"
#include <SPI.h>
#include <EEPROM.h>


FuzzyTank tank; //create the fuzzy tank instance

/*See Fuzzy_Tank_Basic example for notes.*/

void setup()
{
	Serial.begin(115200);
	Serial.println(F("Fuzzy Tank started."));

	tank.begin(); //Required in the setup(). Call this before any other calls to the tank object.

	/*Note: In order for user to override the default settings, the following
	functions need to be placed ***below*** the tank.begin(); statement.
	*/

	/*
	Optional, set the maximum speed for motor output, for forward and backward movements.
	Usable values would be 30~255, where 255 is the fastest.
	Default maximum forward value is 220.
	Default maximum backward value is 160.
	*/
	tank.setMaximumForwardSpeed(255);
	tank.setMaximumBackwardSpeed(255);
	//tank.setMaximumSpeed(255); sets both forward and backward speed

	/*
	Optional, set the smoothness for motor output.
	Higher value yields a smoother motor output change responding
	to joystick command value change, and become more laggy.
	Usable values would be 0~5
	0 means no smoothering, motor output respond to every input change.
	Default value is 2.
	*/
	tank.setMotorSmoothness(2);

	/*
	Optional, set the motor maximum output speed during turning.
	Higher value yields a higher self-rotation rate.
	Usable values would be 30~127
	Maximum value is 127
	Default value is 80
	*/
	tank.setMaximumTurnSpeed(127);

	/*
	Set blood value for each life
	Each normal hit takes 10 blood.
	Default value is 50
	*/
	tank.setBlood(30);

	/*
	Set life value for each game.
	Default value is 3
	*/
	tank.setLife(2);

	/*
	Determinie if tank can be hit after gameover. 
	Set to false if you don't want to be humiliated after gameover.
	Default value is true.
	*/
	tank.canBeHitAfterGameover(true);

	/*
	Set the volume of mp3 player.  Using this method, 
	volume is also stored in the EEPROM, so that different 
	settings can be saved for each board. Users can comment 
	out this line after setting once.
	Note that this code executes after the game has started.
	Default value is 20.
	Maximum value is 30.
	Acceptable value range 1~30.
	*/
	//tank.setVolume(25);

	/*
	Set the status of RF24 connection online/offline voice prompt.
	Default setting is off(false).
	Examples:
	tank.setRF24ConnectionVoicePrompt(true); //Turn the voice prompts on
	tank.setRF24ConnectionVoicePrompt(false); //Turn the voice promtps off
	*/
	tank.setRF24ConnectionVoicePrompt(false);
}

void loop()
{
	tank.update(); //Required in the loop().
}
