#include "Fuzzy_Tank.h"
#include <SPI.h>
#include <EEPROM.h>

/* This example manually sets the RF24 channel number and other information, 
and stores the values in the EEPROM. Each board needs to run this sketch once initially.
This sketch is used on fuzzy tank board only.

Once the board has been setup using the above sketch, the board information
is stored in the EEPROM. The pins will be assigned according to the EEPROM
board version. 
*/


/*Put the desired RF24 channel number here, using 119-5
Currently, each pair of joystick/tank requires 
a dedicated channel. Two pairs in the same channel will cross-talk.
*/
uint8_t userChannel = 117;

/*
Put a device serial number here, not mandatory, just in case 
different code required for different devices. (such as hacks)
Set 0 if there is no special requirements

Edit: this parameter no longer changes the motor related parameters.
*/
uint16_t serialNumber = 0;

/*Put the board number here. This version number will affect
pin setting. Currently the following board versions are supported
- Tank board: V2, (V3), (V4), V5, V6, V7
Note: V3, V4 are hacked version, see change log.
*/
uint8_t boardVersion = 3;

/*Set the motor type. The purpose of the parameter is to set the default speed values.
Different motor type may have different default values.
User can override these default values *** after *** the tank.begin() function.
Valid options are:
MOTOR_TYPE motor = TB16;
MOTOR_TYPE motor = GA20;
*/

MOTOR_TYPE motorType = TB16;

/*Set the gearset type. The purpose of this parameter is to set the correct output direction.
Valid options are:
gearsetType = TOWER_GEARSET_50_TO_1;
gearsetType = SY_PLANETARY_GEARSET_53_TO_1;
gearsetType = SY_PLANETARY_GEARSET_29_TO_1;
*/
GEARSET_TYPE gearsetType = TOWER_GEARSET_50_TO_1;

void setup()
{
	Serial.begin(115200);
	Serial.println("Fuzzy Tank manual setup.");

	//Write the values into EEPROM
	Serial << "Write values into EEPROM...";
	EEPROM.update(TANK_EEPROM_USER_CHANNEL, userChannel);
	EEPROM.update(TANK_EEPROM_SERIAL_NUMBER_HIGH, serialNumber>>8);
	EEPROM.update(TANK_EEPROM_SERIAL_NUMBER_LOW, (uint8_t)serialNumber);
	EEPROM.update(TANK_EEPROM_BOARD_VERSION, boardVersion);
	EEPROM.update(TANK_EEPROM_MOTOR_TYPE, (uint8_t)motorType);
	EEPROM.update(TANK_EEPROM_GEARSET_TYPE, (uint8_t)gearsetType);
	Serial << "done." << endl;
	Serial << endl;

	//read the values from EEPROM to verify the writing
	Serial << "Reading the values from EEPROM..." << endl;
	uint8_t userChannelRead = EEPROM.read(TANK_EEPROM_USER_CHANNEL);
	Serial << "EEPROM Channel = " << userChannelRead << endl;
	uint16_t serialNumberRead = (uint16_t)EEPROM.read(TANK_EEPROM_SERIAL_NUMBER_HIGH) << 8 | (uint16_t)EEPROM.read(TANK_EEPROM_SERIAL_NUMBER_LOW);
	Serial << "Device serial number = " << serialNumberRead << endl;
	uint8_t boardVersionRead = EEPROM.read(TANK_EEPROM_BOARD_VERSION);
	Serial << "EEPROM Board version = " << boardVersionRead << endl;
	motorType = (MOTOR_TYPE)EEPROM.read(TANK_EEPROM_MOTOR_TYPE);
	Serial << F("Motor type = ");
	switch (motorType)
	{
		case DEFAULT_MOTOR:
		{
			Serial << F("default") << endl;
			break;
		}
		case TB16:
		{
			Serial << F("TB16") << endl;
			break;
		}
		case GA20:
		{
			Serial << F("TB16") << endl;
			break;
		}
		default:
		{
			Serial << F("undefined") << endl;
			break;
		}
	}
	gearsetType = (GEARSET_TYPE)EEPROM.read(TANK_EEPROM_GEARSET_TYPE);
	Serial << F("Gearset type = ");
	switch (gearsetType)
	{
		case DEFAULT_GEARSET:
		{
			Serial << F("default") << endl;
			break;
		}
		case TOWER_GEARSET_50_TO_1:
		{
			Serial << F("TOWER_GEARSET_50_TO_1") << endl;
			break;
		}
		case SY_PLANETARY_GEARSET_53_TO_1:
		{
			Serial << F("SY_PLANETARY_GEARSET_53_TO_1") << endl;
			break;
		}
		case SY_PLANETARY_GEARSET_29_TO_1:
		{
			Serial << F("SY_PLANETARY_GEARSET_29_TO_1") << endl;
			break;
		}
		default:
		{
			Serial << F("undefined") << endl;
			break;
		}
	}

	Serial << endl;
	Serial << "Sketch ended." << endl;
}

void loop()
{
}
