/*
 Name:		Fuzzy_Tank.h
 Created:	8/27/2017 2:27:47 AM
 Author:	georgychen
 Editor:	http://www.visualmicro.com

*/


#ifndef _Fuzzy_Tank_h
#define _Fuzzy_Tank_h

#include "arduino.h"
#include <util/delay_basic.h>

#pragma region Common

#pragma region Debug Switches
//#define DEBUG_OUTPUT_CONTROL_DATA
//#define DEBUG_RF24
//#define DEBUG_GAMEPLAY
//#define DEBUG_MOTOR_OUTPUT
//#define DEBUG_MOTOR_MOVEMENT
//#define DEBUG_MP3
//#define DEBUG_IR
#pragma endregion Debug Switches

/*Common Setting*/
#define VERSION_NUMBER "1.2.1"
#define DEFAULT_CHANNEL 120

struct colorRGB {
	uint8_t r;
	uint8_t g;
	uint8_t b;
};


#pragma region Custom RF24
#define PAYLOAD_SIZE 4	//max 32
#define MAX_PAYLOAD_SIZE 32
#define RETRIES_DELAY 4 //delay time in multiples of 250us, at least 4 for low speed
#define RETRIES_COUNT 3
#define RECONNECT_PERIOD 3000
#pragma endregion Custom RF24

#pragma region Orginal Libraries

#pragma region RF24


#ifndef __RF24_CONFIG_H__
#define __RF24_CONFIG_H__

/*** USER DEFINES:  ***/
//#define FAILURE_HANDLING
//#define SERIAL_DEBUG
//#define MINIMAL
//#define SPI_UART  // Requires library from https://github.com/TMRh20/Sketches/tree/master/SPI_UART
//#define SOFTSPI   // Requires library from https://github.com/greiman/DigitalIO

/**********************/
#define rf24_max(a,b) (a>b?a:b)
#define rf24_min(a,b) (a<b?a:b)

#if defined SPI_HAS_TRANSACTION && !defined SPI_UART && !defined SOFTSPI
#define RF24_SPI_TRANSACTIONS
#endif


//ATXMega
#if defined(__AVR_ATxmega64D3__) || defined(__AVR_ATxmega128D3__) || defined(__AVR_ATxmega192D3__) || defined(__AVR_ATxmega256D3__) || defined(__AVR_ATxmega384D3__) // In order to be available both in windows and linux this should take presence here.
#define XMEGA
#define XMEGA_D3
#include "utility/ATXMegaD3/RF24_arch_config.h"
#elif ( !defined (ARDUINO) ) // Any non-arduino device is handled via configure/Makefile

// The configure script detects device and copies the correct includes.h file to /utility/includes.h
// This behavior can be overridden by calling configure with respective parameters
// The includes.h file defines either RF24_RPi, MRAA, LITTLEWIRE or RF24_SPIDEV and includes the correct RF24_arch_config.h file
#include "utility/includes.h"

//ATTiny  
#elif defined(__AVR_ATtiny25__) || defined(__AVR_ATtiny45__) || defined(__AVR_ATtiny85__) || defined(__AVR_ATtiny24__) || defined(__AVR_ATtiny44__) || defined(__AVR_ATtiny84__) || defined(__AVR_ATtiny2313__) || defined(__AVR_ATtiny4313__) || defined(__AVR_ATtiny861__)  
#define RF24_TINY
#include "utility/ATTiny/RF24_arch_config.h"


//LittleWire  
#elif defined(LITTLEWIRE)

#include "utility/LittleWire/RF24_arch_config.h"

//Teensy  
#elif defined (TEENSYDUINO)

#include "utility/Teensy/RF24_arch_config.h"  
//Everything else
#else 

#include <Arduino.h>

// RF modules support 10 Mhz SPI bus speed
const uint32_t RF24_SPI_SPEED = 10000000;

#if defined (ARDUINO) && !defined (__arm__) && !defined (__ARDUINO_X86__)
#if defined SPI_UART
#include <SPI_UART.h>
#define _SPI uspi
#elif defined SOFTSPI
// change these pins to your liking
//
const uint8_t SOFT_SPI_MISO_PIN = 16;
const uint8_t SOFT_SPI_MOSI_PIN = 15;
const uint8_t SOFT_SPI_SCK_PIN = 14;
const uint8_t SPI_MODE = 0;
#define _SPI spi

#else	    
#include <SPI.h>
#define _SPI SPI
#endif
#else
// Define _BV for non-Arduino platforms and for Arduino DUE
#include <stdint.h>
#include <stdio.h>
#include <string.h>


#if defined(__arm__) || defined (__ARDUINO_X86__)
#if defined (__arm__) && defined (SPI_UART)
#include <SPI_UART.h>
#define _SPI uspi
#else
#include <SPI.h>
#define _SPI SPI
#endif
#elif !defined(__arm__) && !defined (__ARDUINO_X86__)
extern HardwareSPI SPI;
#endif

#define _BV(x) (1<<(x))

#endif

#ifdef SERIAL_DEBUG
#define IF_SERIAL_DEBUG(x) ({x;})
#else
#define IF_SERIAL_DEBUG(x)
#if defined(RF24_TINY)
#define printf_P(...)
#endif
#endif

#if  defined (__ARDUINO_X86__)
#define printf_P printf
#define _BV(bit) (1<<(bit))
#endif

// Progmem is Arduino-specific
// Arduino DUE is arm and does not include avr/pgmspace
#if defined (ARDUINO_ARCH_ESP8266)

#define PSTR(x) (x)
#define printf Serial.printf
#define sprintf(...) os_sprintf( __VA_ARGS__ )
#define printf_P printf
#define strlen_P strlen  
#define PROGMEM
#define pgm_read_byte(addr) (*(const unsigned char *)(addr))
#define pgm_read_word(p) (*(p))
#define PRIPSTR "%s"

#elif defined(ARDUINO) && ! defined(__arm__) && !defined (__ARDUINO_X86__) || defined(XMEGA)
#include <avr/pgmspace.h>
#define PRIPSTR "%S"
#else
#if ! defined(ARDUINO) // This doesn't work on Arduino DUE
typedef char const char;
#else // Fill in pgm_read_byte that is used, but missing from DUE
#define pgm_read_byte(addr) (*(const unsigned char *)(addr))
#endif


typedef uint16_t prog_uint16_t;
#define PSTR(x) (x)
#define printf_P printf
#define strlen_P strlen
#define PROGMEM
#define pgm_read_word(p) (*(p))

#define PRIPSTR "%s"

#endif

#endif



#endif // __RF24_CONFIG_H__

#ifndef __nRF24L01_MEMORY_MAP_
#define __nRF24L01_MEMORY_MAP_
/* Memory Map */
#define NRF_CONFIG      0x00
#define EN_AA       0x01
#define EN_RXADDR   0x02
#define SETUP_AW    0x03
#define SETUP_RETR  0x04
#define RF_CH       0x05
#define RF_SETUP    0x06
#define NRF_STATUS  0x07
#define OBSERVE_TX  0x08
#define CD          0x09
#define RX_ADDR_P0  0x0A
#define RX_ADDR_P1  0x0B
#define RX_ADDR_P2  0x0C
#define RX_ADDR_P3  0x0D
#define RX_ADDR_P4  0x0E
#define RX_ADDR_P5  0x0F
#define TX_ADDR     0x10
#define RX_PW_P0    0x11
#define RX_PW_P1    0x12
#define RX_PW_P2    0x13
#define RX_PW_P3    0x14
#define RX_PW_P4    0x15
#define RX_PW_P5    0x16
#define FIFO_STATUS 0x17
#define DYNPD	    0x1C
#define FEATURE	    0x1D

/* Bit Mnemonics */
#define MASK_RX_DR  6
#define MASK_TX_DS  5
#define MASK_MAX_RT 4
#define EN_CRC      3
#define CRCO        2
#define PWR_UP      1
#define PRIM_RX     0
#define ENAA_P5     5
#define ENAA_P4     4
#define ENAA_P3     3
#define ENAA_P2     2
#define ENAA_P1     1
#define ENAA_P0     0
#define ERX_P5      5
#define ERX_P4      4
#define ERX_P3      3
#define ERX_P2      2
#define ERX_P1      1
#define ERX_P0      0
#define AW          0
#define ARD         4
#define ARC         0
#define PLL_LOCK    4
#define RF_DR       3
#define RF_PWR      6
#define RX_DR       6
#define TX_DS       5
#define MAX_RT      4
#define RX_P_NO     1
#define TX_FULL     0
#define PLOS_CNT    4
#define ARC_CNT     0
#define TX_REUSE    6
#define FIFO_FULL   5
#define TX_EMPTY    4
#define RX_FULL     1
#define RX_EMPTY    0
#define DPL_P5	    5
#define DPL_P4	    4
#define DPL_P3	    3
#define DPL_P2	    2
#define DPL_P1	    1
#define DPL_P0	    0
#define EN_DPL	    2
#define EN_ACK_PAY  1
#define EN_DYN_ACK  0

/* Instruction Mnemonics */
#define R_REGISTER    0x00
#define W_REGISTER    0x20
#define REGISTER_MASK 0x1F
#define ACTIVATE      0x50
#define R_RX_PL_WID   0x60
#define R_RX_PAYLOAD  0x61
#define W_TX_PAYLOAD  0xA0
#define W_ACK_PAYLOAD 0xA8
#define FLUSH_TX      0xE1
#define FLUSH_RX      0xE2
#define REUSE_TX_PL   0xE3
#define NOP           0xFF

/* Non-P omissions */
#define LNA_HCURR   0

/* P model memory Map */
#define RPD         0x09
#define W_TX_PAYLOAD_NO_ACK  0xB0

/* P model bit Mnemonics */
#define RF_DR_LOW   5
#define RF_DR_HIGH  3
#define RF_PWR_LOW  1
#define RF_PWR_HIGH 2

#endif //__nRF24L01_MEMORY_MAP_

/**
* Power Amplifier level.
*
* For use with setPALevel()
*/
typedef enum { RF24_PA_MIN = 0, RF24_PA_LOW, RF24_PA_HIGH, RF24_PA_MAX, RF24_PA_ERROR } rf24_pa_dbm_e;

/**
* Data rate.  How fast data moves through the air.
*
* For use with setDataRate()
*/
typedef enum { RF24_1MBPS = 0, RF24_2MBPS, RF24_250KBPS } rf24_datarate_e;

/**
* CRC Length.  How big (if any) of a CRC is included.
*
* For use with setCRCLength()
*/
typedef enum { RF24_CRC_DISABLED = 0, RF24_CRC_8, RF24_CRC_16 } rf24_crclength_e;

#pragma endregion RF24

#pragma region IR
#ifndef IRLREMOTE_H
#define IRLREMOTE_H
#define IRL_BLOCKING true


/*
Copyright (c) 2014-2015 NicoHood
See the readme for credit to other people.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/



// software version
#define IRL_VERSION 174

#include <Arduino.h>

// missing 1.0.6 definition workaround
#ifndef NOT_AN_INTERRUPT
#define NOT_AN_INTERRUPT -1
#endif
#define  IRLREMOTE_AVR

// delay_basic is only for avrs. with ARM sending is currently not possible
#ifdef IRLREMOTE_AVR
#include <util/delay_basic.h>
#endif

//================================================================================
// Definitions
//================================================================================

//NEC
//IRP notation: {38.4k,564}<1,-1|1,-3>(16,-8,D:8,S:8,F:8,~F:8,1,-78,(16,-4,1,-173)*)
// Lead + Space logic
#define NEC_HZ 38400
#define NEC_PULSE 564UL
#define NEC_BLOCKS 4
#define NEC_ADDRESS_LENGTH 16
#define NEC_COMMAND_LENGTH 16
#define NEC_LENGTH 2 + NEC_BLOCKS*8*2 // 2 for lead + space, each block has 8bits: mark and space
#define NEC_TIMEOUT NEC_PULSE*173
#define NEC_MARK_LEAD NEC_PULSE*16
#define NEC_SPACE_LEAD NEC_PULSE*8
#define NEC_SPACE_HOLDING NEC_PULSE*4
#define NEC_MARK_ZERO NEC_PULSE*1
#define NEC_MARK_ONE NEC_PULSE*1
#define NEC_SPACE_ZERO NEC_PULSE*1
#define NEC_SPACE_ONE NEC_PULSE*3


typedef enum IRType {
	IR_NO_PROTOCOL, // 0
	IR_USER, // 1
	IR_ALL, // 2
	IR_NEC, // ...
};

// definitions to use decoding functions with extra accuracy
#define IR_EXTRA_ACCURACY true
#define IR_NO_EXTRA_ACCURACY false

// definition for sending to determine what should be send first
#define IR_ADDRESS_FIRST true
#define IR_COMMAND_FIRST false

// definition to convert an uint8_t array to an uint16_t/uint32_t at any position (thx timeage!)
#define UINT16_AT_OFFSET(p_to_8, offset)    ((uint16_t)*((const uint16_t *)((p_to_8)+(offset))))
#define UINT32_AT_OFFSET(p_to_8, offset)    ((uint32_t)*((const uint32_t *)((p_to_8)+(offset))))

// definition to get the higher value
#define MAX(a,b) \
   ({ __typeof__ (a) _a = (a); \
       __typeof__ (b) _b = (b); \
     _a > _b ? _a : _b; })

// definition to get the lower value
#define MIN(a,b) \
   ({ __typeof__ (a) _a = (a); \
       __typeof__ (b) _b = (b); \
     _a < _b ? _a : _b; })

//================================================================================
// Prototypes
//================================================================================

// function called on a valid IR event, must be overwritten by the user
void __attribute__((weak)) IREvent(uint8_t protocol, uint16_t address, uint32_t command);

// optinal user function called when the user wants to implement its own decoding function
void __attribute__((weak)) decodeIR(const uint16_t duration);

// called by interrupt CHANGE
template <IRType irType>
inline void IRLinterrupt(void) __attribute__((always_inline));

// special decode function for each protocol
inline void decodeAll(const uint16_t duration) __attribute__((always_inline));
template <bool extraAccuracy> inline void decodeNec(const uint16_t duration);


// functions to check if the received data is valid with the protocol checksums
inline bool IRLcheckInverse0(uint8_t data[]) __attribute__((always_inline));
inline bool IRLcheckInverse1(uint8_t data[]) __attribute__((always_inline));
inline bool IRLcheckHolding(uint8_t data[]) __attribute__((always_inline));
inline bool IRLcheckXOR0(uint8_t data[]) __attribute__((always_inline));

// default decoder helper function
template <uint8_t irLength, uint16_t timeoutThreshold, uint16_t markLeadThreshold, uint16_t spaceLeadThreshold,
	uint16_t spaceLeadHoldingThreshold, uint16_t markThreshold, uint16_t spaceThreshold,
	uint16_t markTimeout, uint16_t spaceTimeout>
	inline bool IRLdecode(uint16_t duration, uint8_t data[]) __attribute__((always_inline));

// functions to send the protocol
template <IRType irType>
inline void IRLwrite(const uint8_t pin, uint16_t address, uint32_t command) __attribute__((always_inline));

template <uint8_t addressLength, uint8_t commandLength,
	uint16_t Hz, bool addressFirst,
	uint16_t markLead, uint16_t spaceLead,
	uint16_t markZero, uint16_t markOne,
	uint16_t spaceZero, uint16_t spaceOne>
	inline void IRLsend(volatile uint8_t * outPort, uint8_t bitmask, uint16_t address, uint32_t command) __attribute__((always_inline));

// functions to set the pin high or low (with bitbang pwm)
inline void IRLmark(const uint16_t Hz, volatile uint8_t * outPort, uint8_t bitMask, uint16_t time) __attribute__((always_inline));
inline void IRLspace(volatile uint8_t * outPort, uint8_t bitMask, uint16_t time) __attribute__((always_inline));

//================================================================================
// Inline Implementations (receive)
//================================================================================

extern uint32_t IRL_LastTime;

template <IRType irType>
void IRLinterrupt(void) {
	//save the duration between the last reading
	uint32_t time = micros();
	uint32_t duration_32 = time - IRL_LastTime;
	IRL_LastTime = time;

	// calculate 16 bit duration. On overflow sets duration to a clear timeout
	uint16_t duration = 0xFFFF;
	if (duration_32 <= 0xFFFF)
		duration = duration_32;

	// determinate which decode function must be called
	switch (irType) {
	case IR_USER:
		// userfunction that can be overwritten
		decodeIR(duration);
		break;
	case IR_ALL:
		decodeAll(duration);
		break;
	case IR_NEC:
		decodeNec<IR_NO_EXTRA_ACCURACY>(duration);
		break;
	}
}

void decodeAll(const uint16_t duration) {
	// go through all known protocols and decode with more (resource unfriendly) accuration
	decodeNec<IR_EXTRA_ACCURACY>(duration);
}

template <bool extraAccuracy> void decodeNec(const uint16_t duration) {
	// temporary buffer to hold bytes for decoding this protocol
	static uint8_t data[NEC_BLOCKS];

	// pass the duration to the decoding function
	bool newInput;
	if (extraAccuracy) // no accuracy set at the moment, no conflict detected yet
		newInput = IRLdecode <NEC_LENGTH, (NEC_TIMEOUT + NEC_MARK_LEAD) / 2, // irLength, timeoutThreshold
		(NEC_MARK_LEAD + NEC_SPACE_ONE) / 2, (NEC_SPACE_LEAD + NEC_SPACE_HOLDING) / 2, // markLeadThreshold, spaceLeadThreshold
		(NEC_SPACE_HOLDING + NEC_SPACE_ONE) / 2, 0, // spaceLeadHoldingThreshold, markThreshold
		(NEC_SPACE_ONE + NEC_SPACE_ZERO) / 2, // spaceThreshold
		0, 0>// markTimeout, spaceTimeout
		(duration, data);
	else
		newInput = IRLdecode <NEC_LENGTH, (NEC_TIMEOUT + NEC_MARK_LEAD) / 2, // irLength, timeoutThreshold
		(NEC_MARK_LEAD + NEC_SPACE_ONE) / 2, (NEC_SPACE_LEAD + NEC_SPACE_HOLDING) / 2, // markLeadThreshold, spaceLeadThreshold
		(NEC_SPACE_HOLDING + NEC_SPACE_ONE) / 2, 0, // spaceLeadHoldingThreshold, markThreshold
		(NEC_SPACE_ONE + NEC_SPACE_ZERO) / 2, // spaceThreshold
		0, 0>// markTimeout, spaceTimeout
		(duration, data);

	if (newInput) {
		// Check if the protcol's checksum is correct
		bool holding;
		if (IRLcheckInverse1(data) || (holding = IRLcheckHolding(data))) {
			// normally NEC also check for the inverse of the address.
			// newer remotes dont have this because of the wide used protocol all addresses were already used
			// to make it less complicated its left out and the user can check the command inverse himself if needed
			//if (!holding && !extended && !IRLcheckInverse0(data))
			//	return;

			// save address + command and trigger event
			uint16_t address = UINT16_AT_OFFSET(data, 0);
			uint32_t command = UINT16_AT_OFFSET(data, 2);
			IREvent(IR_NEC, address, command);
		}
	}
}


bool IRLcheckInverse0(uint8_t data[]) {
	// check if byte 2 and is the inverse of byte 3
	if (uint8_t((data[0] ^ (~data[1]))) == 0)
		return true;
	else return false;
}

bool IRLcheckInverse1(uint8_t data[]) {
	// check if byte 0 and is the inverse of byte 1
	if (uint8_t((data[2] ^ (~data[3]))) == 0)
		return true;
	else return false;
}

bool IRLcheckHolding(uint8_t data[]) {
	// check if at least the command is always 1
	if (data[2] == 0xFF && data[3] == 0xFF)
		return true;
	else return false;
}

//bool IRLcheckXOR0(uint8_t data[]) {
//	// this function is used for panasonic checksum
//	if (uint8_t(data[2] ^ data[3] ^ data[4]) == data[5])
//		return true;
//	else return false;
//}

// multifunctional template for receiving
template <uint8_t irLength, uint16_t timeoutThreshold, uint16_t markLeadThreshold, uint16_t spaceLeadThreshold,
	uint16_t spaceLeadHoldingThreshold, uint16_t markThreshold, uint16_t spaceThreshold,
	uint16_t markTimeout, uint16_t spaceTimeout>
	bool IRLdecode(uint16_t duration, uint8_t data[]) {

	// variables for ir processing
	static uint8_t count = 0;

	// if timeout always start next possible reading and abort any pending readings
	if (duration >= timeoutThreshold)
		count = 1;

	// on a reset (error in decoding) we are waiting for a timeout to start a new reading again
	// this is to not conflict with other protocols while they are sending 0/1 which might be similar to a lead in another protocol
	else if (count == 0)
		return false;

	// check pulses for mark/space and lead + logical 0/1 seperate
	else {
		// Mark pulses (odd numbers)
		if (count % 2 == 1) {
			// check Mark Lead (needs a timeout or a correct signal)
			if (markLeadThreshold && count == 1) {
				// lead is okay// wrong lead
				if (duration <= markLeadThreshold) {
					count = 0;
					return false;
				}
			}

			else {
				// check for timeout if needed (might be a different protocol)
				if (markTimeout && duration > markTimeout) {
					count = 0;
					return false;
				}

				// only check values if the protocol has different logical space pulses
				else if (markThreshold) {

					// get number of the Mark Bits (starting from zero)
					uint8_t length;
					// only save every 2nd value, substract the first two lead pulses
					if (!spaceThreshold)
						length = (count / 2) - 1;
					// special case: spaces and marks both have data in the pulse
					else length = count - 2;

					// move bits and write 1 or 0 depending on the duration
					// 1.7: changed from MSB to LSB. somehow takes a bit more flash but is correct and easier to handle.
					data[length / 8] >>= 1;
					if (duration > markThreshold)
						data[length / 8] |= 0x80;
					//else // normally not needed through the bitshift
					//	data[length / 8] &= ~0x80;
				}

				// check last input (always a mark)
				if (count > irLength) {
					count = 0;
					return true;
				}
			}
		}

		// Space pulses (even numbers)
		else {
			//check Space Lead/Space Holding
			if (spaceLeadThreshold && count == 2) {
				// normal Space, next reading
				if (duration > spaceLeadThreshold);

				// Button holding (if supported by protocol)
				else if (spaceLeadHoldingThreshold && duration > spaceLeadHoldingThreshold) {
					// set command to 0xFF if button is held down
					if (irLength == (4 * 8 * 2 + 2)) {
						data[0] = data[1] = 0x00;
						data[2] = data[3] = 0xFF;
					}
					count = 0;
					return true;
				}
				// wrong space
				else {
					count = 0;
					return false;
				}
			}
			else {
				// check for timeout if needed (might be a different protocol)
				if (spaceTimeout && duration > spaceTimeout) {
					count = 0;
					return false;
				}

				// only check values if the protocol has different logical space pulses
				else if (spaceThreshold) {

					// get number of the Space Bits (starting from zero)
					uint8_t length;
					// only save every 2nd value, substract the first two lead pulses
					if (!markThreshold)
						length = (count / 2) - 2;
					// special case: spaces and marks both have data in the pulse
					else length = count - 2;

					// move bits and write 1 or 0 depending on the duration
					// 1.7: changed from MSB to LSB. somehow takes a bit more flash but is correct and easier to handle.
					data[length / 8] >>= 1;
					if (duration > spaceThreshold)
						data[length / 8] |= 0x80;
					//else // normally not needed through the bitshift
					//	data[length / 8] &= ~0x80;
				}
			}
		}

		// next reading, no errors
		count++;
	}

	// no valid input (yet)
	return false;
}

//================================================================================
// Inline Implementations (send)
//================================================================================

template <IRType irType>
void IRLwrite(const uint8_t pin, uint16_t address, uint32_t command)
{
	// get the port mask and the pointers to the out/mode registers for faster access
	uint8_t bitMask = digitalPinToBitMask(pin);
	uint8_t port = digitalPinToPort(pin);
	volatile uint8_t * outPort = portOutputRegister(port);
	volatile uint8_t * modePort = portModeRegister(port);

	// set pin to OUTPUT and LOW
	*modePort |= bitMask;
	*outPort &= ~bitMask;

	// disable interrupts
	uint8_t oldSREG = SREG;
	cli();

	switch (irType) {

	case IR_NEC:
		// NEC only sends the data once
		if (command == 0xFFF)
			// send holding indicator
			IRLsend<0, 0, NEC_HZ, 0, NEC_MARK_LEAD, NEC_SPACE_HOLDING,
			0, 0, 0, 0>
			(outPort, bitMask, address, command);
		else
			// send data
			IRLsend<NEC_ADDRESS_LENGTH, NEC_COMMAND_LENGTH, NEC_HZ, IR_ADDRESS_FIRST, NEC_MARK_LEAD, NEC_SPACE_LEAD,
			NEC_MARK_ZERO, NEC_MARK_ONE, NEC_SPACE_ZERO, NEC_SPACE_ONE>
			(outPort, bitMask, address, command);
		break;
	}

	// enable interrupts
	SREG = oldSREG;

	// set pin to INPUT again to be save
	*modePort &= ~bitMask;
}

// multifunctional template for sending
template <uint8_t addressLength, uint8_t commandLength,
	uint16_t Hz, bool addressFirst,
	uint16_t markLead, uint16_t spaceLead,
	uint16_t markZero, uint16_t markOne,
	uint16_t spaceZero, uint16_t spaceOne>
	void IRLsend(volatile uint8_t * outPort, uint8_t bitmask, uint16_t address, uint32_t command) {

	// send lead mark
	if (markLead)
		IRLmark(Hz, outPort, bitmask, markLead);

	// send lead space
	if (spaceLead)
		IRLspace(outPort, bitmask, spaceLead);

	// abort if input values are both the same (sending a holding signal for example)
	if (markOne != markZero || spaceOne != spaceZero)
	{
		// go through all bits
		for (uint8_t i = 0; i < (addressLength + commandLength); i++) {
			// determine if its a logical one or zero, starting with address or command
			bool bitToSend;
			if (addressFirst && i < addressLength || !addressFirst && i >= commandLength) {
				bitToSend = address & 0x01;
				address >>= 1;
			}
			else {
				bitToSend = command & 0x01;
				command >>= 1;
			}

			// send logic mark bits if needed
			if (markOne != markZero) {
				// modulate if spaces dont have logic, else only every even number
				if (spaceOne == spaceZero || spaceOne != spaceZero && i % 2 == 0)
				{
					if (bitToSend)
						IRLmark(Hz, outPort, bitmask, markOne);
					else
						IRLmark(Hz, outPort, bitmask, markZero);
				}
			}
			else
				IRLmark(Hz, outPort, bitmask, markOne);

			// send logic space bits if needed
			if (spaceOne != spaceZero) {
				// modulate if marks dont have logic, else only every odd number
				if (markOne == markZero || markOne != markZero && i % 2 == 1) {
					if (bitToSend)
						IRLspace(outPort, bitmask, spaceOne);
					else
						IRLspace(outPort, bitmask, spaceZero);
				}
			}
			else
				IRLspace(outPort, bitmask, spaceOne);
		}
	}

	// finish mark
	IRLmark(Hz, outPort, bitmask, markZero); //TODO zero or one for sony?
	IRLspace(outPort, bitmask, 0);
}

void IRLmark(const uint16_t Hz, volatile uint8_t * outPort, uint8_t bitMask, uint16_t time) {
	/*
	Bitbangs PWM in the given Hz number for the given time
	________________________________________________________________________________
	Delay calculation:
	F_CPU/1.000.000 to get number of cycles/uS
	/3 to get the number of loops needed for 1ms (1loop = 3 cycles)

	Multiply with the number of ms delay:
	1/kHz to get the seconds
	* 1.000.000 to get it in uS
	/2 to get half of a full pulse

	Substract the while, portmanipulation, loop overhead /3 loop cycles

	F_CPU(16.000.000)            1 * 1.000.000(pulse in uS)   12(overhead)
	========================== * ========================== - ==============
	1.000.000 * 3(loop cycles)   Hz * 2(half of a pulse)      3(loop cycles)

	<==>

	F_CPU(16.000.000) - (12(overhead) * Hz * 2(half of a pulse))
	===========================================================
	Hz * 2(half of a on/off pulse) * 3(loop cycles)

	________________________________________________________________________________
	Iterations calculation:
	Devide time with cycles in while loop
	Multiply this with the cycles per uS
	cycles per while loop: 3(loop cycles) * delay + overhead

	time * (F_CPU(16.000.000) / 1.000.000)
	======================================
	delay*3(loop cycles) + overhead
	*/

	const uint32_t loopCycles = 3;
	const uint32_t overHead = 12; // just a guess from try + error
	uint8_t delay = (F_CPU - (overHead * Hz * 2UL)) / (Hz * 2UL * loopCycles);
	uint16_t iterations = (time*(F_CPU / 1000000UL)) / (delay * loopCycles + overHead);

	while (iterations--) {
		// flip pin state and wait for the calculated time
		*outPort ^= bitMask;
#ifdef IRLREMOTE_AVR
		_delay_loop_1(delay);
#else

#endif
	}
}

void IRLspace(volatile uint8_t * outPort, uint8_t bitMask, uint16_t time) {
	// Sends an IRLspace for the specified number of microseconds.
	// A IRLspace is no output, so the PWM output is disabled.
	*outPort &= ~bitMask; // write pin LOW
	delayMicroseconds(time);
	//_delay_loop_2(time*(F_CPU/1000000UL)/4UL);
}

#endif
#pragma endregion IR

#pragma region STREAMING_LIBRARY
#define STREAMING_LIBRARY_VERSION 5

// Generic template
template<class T>
inline Print &operator <<(Print &stream, T arg)
{
	stream.print(arg); return stream;
}

struct _BASED
{
	long val;
	int base;
	_BASED(long v, int b) : val(v), base(b)
	{}
};

#if ARDUINO >= 100

struct _BYTE_CODE
{
	byte val;
	_BYTE_CODE(byte v) : val(v)
	{}
};
#define _BYTE(a)    _BYTE_CODE(a)

inline Print &operator <<(Print &obj, const _BYTE_CODE &arg)
{
	obj.write(arg.val); return obj;
}

#else

#define _BYTE(a)    _BASED(a, BYTE)

#endif

#define _HEX(a)     _BASED(a, HEX)
#define _DEC(a)     _BASED(a, DEC)
#define _OCT(a)     _BASED(a, OCT)
#define _BIN(a)     _BASED(a, BIN)

// Specialization for class _BASED
// Thanks to Arduino forum user Ben Combee who suggested this 
// clever technique to allow for expressions like
//   Serial << _HEX(a);

inline Print &operator <<(Print &obj, const _BASED &arg)
{
	obj.print(arg.val, arg.base); return obj;
}

#if ARDUINO >= 18
// Specialization for class _FLOAT
// Thanks to Michael Margolis for suggesting a way
// to accommodate Arduino 0018's floating point precision
// feature like this:
//   Serial << _FLOAT(gps_latitude, 6); // 6 digits of precision

struct _FLOAT
{
	float val;
	int digits;
	_FLOAT(double v, int d) : val(v), digits(d)
	{}
};

inline Print &operator <<(Print &obj, const _FLOAT &arg)
{
	obj.print(arg.val, arg.digits); return obj;
}
#endif

// Specialization for enum _EndLineCode
// Thanks to Arduino forum user Paul V. who suggested this
// clever technique to allow for expressions like
//   Serial << "Hello!" << endl;

enum _EndLineCode { endl };

inline Print &operator <<(Print &obj, _EndLineCode arg)
{
	obj.println(); return obj;
}
#pragma endregion STREAMING_LIBRARY

#pragma region WS2812
// Timing in ns
#define w_zeropulse   350
#define w_onepulse    900
#define w_totalperiod 1250

// Fixed cycles used by the inner loop
#define w_fixedlow    3
#define w_fixedhigh   6
#define w_fixedtotal  10   

// Insert NOPs to match the timing, if possible
#define w_zerocycles    (((F_CPU/1000)*w_zeropulse          )/1000000)
#define w_onecycles     (((F_CPU/1000)*w_onepulse    +500000)/1000000)
#define w_totalcycles   (((F_CPU/1000)*w_totalperiod +500000)/1000000)

// w1 - nops between rising edge and falling edge - low
#define w1 (w_zerocycles-w_fixedlow)
// w2   nops between fe low and fe high
#define w2 (w_onecycles-w_fixedhigh-w1)
// w3   nops to complete loop
#define w3 (w_totalcycles-w_fixedtotal-w1-w2)

#if w1>0
#define w1_nops w1
#else
#define w1_nops  0
#endif

// The only critical timing parameter is the minimum pulse length of the "0"
// Warn or throw error if this timing can not be met with current F_CPU settings.
#define w_lowtime ((w1_nops+w_fixedlow)*1000000)/(F_CPU/1000)
#if w_lowtime>550
#error "Light_ws2812: Sorry, the clock speed is too low. Did you set F_CPU correctly?"
#elif w_lowtime>450
#warning "Light_ws2812: The timing is critical and may only work on WS2812B, not on WS2812(S)."
#warning "Please consider a higher clockspeed, if possible"
#endif   

#if w2>0
#define w2_nops w2
#else
#define w2_nops  0
#endif

#if w3>0
#define w3_nops w3
#else
#define w3_nops  0
#endif

#define w_nop1  "nop      \n\t"
#define w_nop2  "rjmp .+0 \n\t"
#define w_nop4  w_nop2 w_nop2
#define w_nop8  w_nop4 w_nop4
#define w_nop16 w_nop8 w_nop8
#pragma endregion WS2812

#pragma endregion Orginal Libraries

#pragma endregion Common


#pragma region Tank

#pragma region Tank Setup and pairing

/*Tank Setting*/
#define TANK_EEPROM_SERIAL_NUMBER_HIGH 0X01	
#define TANK_EEPROM_SERIAL_NUMBER_LOW 0x02
#define TANK_EEPROM_USER_CHANNEL 0x03
#define	TANK_EEPROM_BOARD_VERSION 0x04
#define TANK_EEPROM_MOTOR_TYPE 0x05
#define TANK_EEPROM_GEARSET_TYPE 0x06
#define TANK_EEPROM_HEAD_COLOR_INDEX 0x07
#define TANK_EEPROM_EYES_COLOR_INDEX 0x08
#define TANK_EEPROM_MP3_VOLUME 0x09

/*
EEPROM_MOTOR_TYPE
value	255	undefined
0	default motor
1	TB16
2	GA20

EEPROM_GEARSET_TYPE
value	255	undefined
0	default gearset
1	tower gear set 50:1
2	shayangye planetary gear set 53:1
3	shayangye planetary gear set 29:1

todo: different combination result in different default values
*/

typedef enum MOTOR_TYPE
{
	DEFAULT_MOTOR,
	TB16,
	GA20,
};

typedef enum GEARSET_TYPE
{
	DEFAULT_GEARSET,
	TOWER_GEARSET_50_TO_1,
	SY_PLANETARY_GEARSET_53_TO_1,
	SY_PLANETARY_GEARSET_29_TO_1,
};
#pragma endregion Tank Setup and pairing

#pragma region Tank Pin assignment

/*
To add a supported board
(1)Copy and modify this #define section, with new _V? suffix
(2)Append the array uint8_t validBoardVersions[4] = { 2,5,6,7 }; with the new board version
(3)Add a case ? section in void FuzzyTank::setPins()
(4)Increase the value of SUPPORTED_BOARD_TYPES
*/
#define TANK_SUPPORTED_BOARD_TYPES 6  
//====================== V2 board ========================
#define TANK_YX5200_SERIAL_PIN_V2 7	//This is the arduino software serial TX pin
#define TANK_MOTOR_A_PWM_PIN_V2 5
#define TANK_MOTOR_A_IN1_PIN_V2 A4
#define TANK_MOTOR_A_IN2_PIN_V2 A3
#define TANK_MOTOR_B_PWM_PIN_V2 6
#define TANK_MOTOR_B_IN1_PIN_V2 A2
#define TANK_MOTOR_B_IN2_PIN_V2 A1
#define TANK_RF24_CSN_PIN_V2 A5		//A7 cannot be used as digital output, hacked to A5
#define TANK_RF24_CE_PIN_V2 21		//use the A7(D21) as dummy pin
#define TANK_IR_RECEIVER_PIN_V2 2
#define TANK_IR_SENDER_PIN_V2 3
#define TANK_WS2812_LED_PIN_V2 4		//ws2812 rgb led
#define TANK_FRONT_LED_PIN_V2 8		//A 1w led (visible light)
#define TANK_SERVO1_PIN_V2 9			//tico servo
//====================== V3 hack board ========================
//This is a hack board from V7, D5<->D9, D6<->D10, to test pwm input from timer1
//tico servo at D9 removed
#define TANK_YX5200_SERIAL_PIN_V3 7  //This is the arduino software serial TX pin
#define TANK_MOTOR_A_PWM_PIN_V3 9
#define TANK_MOTOR_A_IN1_PIN_V3 A0
#define TANK_MOTOR_A_IN2_PIN_V3 8
#define TANK_MOTOR_B_PWM_PIN_V3 10
#define TANK_MOTOR_B_IN1_PIN_V3 A2
#define TANK_MOTOR_B_IN2_PIN_V3 A1
#define TANK_RF24_CSN_PIN_V3 A3
#define TANK_RF24_CE_PIN_V3 21		//use the A7(D21) as dummy pin
#define TANK_IR_RECEIVER_PIN_V3 2
#define TANK_IR_SENDER_PIN_V3 3
#define TANK_WS2812_LED_PIN_V3 4		//ws2812 rgb led
#define TANK_FRONT_LED_PIN_V3 6		//A 1w led (visible light)
#define TANK_SERVO1_PIN_V3 5			//bit bang servo
//====================== V4 hack board ========================
//This is a hack board from V6, D5<->D9, D6<->D10, to test pwm input from timer1
//tico servo at D9 removed
#define TANK_YX5200_SERIAL_PIN_V4 7  //This is the arduino software serial TX pin
#define TANK_MOTOR_A_PWM_PIN_V4 9	//5->9
#define TANK_MOTOR_A_IN1_PIN_V4 A0
#define TANK_MOTOR_A_IN2_PIN_V4 6	//10->6
#define TANK_MOTOR_B_PWM_PIN_V4 10	//6->10
#define TANK_MOTOR_B_IN1_PIN_V4 A2
#define TANK_MOTOR_B_IN2_PIN_V4 A1
#define TANK_RF24_CSN_PIN_V4 A3
#define TANK_RF24_CE_PIN_V4 21		//use the A7(D21) as dummy pin
#define TANK_IR_RECEIVER_PIN_V4 2
#define TANK_IR_SENDER_PIN_V4 3
#define TANK_WS2812_LED_PIN_V4 4		//ws2812 rgb led
#define TANK_FRONT_LED_PIN_V4 8		//A 1w led (visible light)
#define TANK_SERVO1_PIN_V4 5			//bit bang servo
//====================== V5 board ========================
#define TANK_YX5200_SERIAL_PIN_V5 7	//This is the arduino software serial TX pin
#define TANK_MOTOR_A_PWM_PIN_V5 5
#define TANK_MOTOR_A_IN1_PIN_V5 A6	//A6 cannot be used as digital output
#define TANK_MOTOR_A_IN2_PIN_V5 10
#define TANK_MOTOR_B_PWM_PIN_V5 6
#define TANK_MOTOR_B_IN1_PIN_V5 A2		
#define TANK_MOTOR_B_IN2_PIN_V5 A1
#define TANK_RF24_CSN_PIN_V5 A5
#define TANK_RF24_CE_PIN_V5 21		//use the A7(D21) as dummy pin
#define TANK_IR_RECEIVER_PIN_V5 2
#define TANK_IR_SENDER_PIN_V5 3
#define TANK_WS2812_LED_PIN_V5 4		//ws2812 rgb led
#define TANK_FRONT_LED_PIN_V5 8		//A 1w led (visible light)
#define TANK_SERVO1_PIN_V5 9			//tico servo
//====================== V6 board ========================
#define TANK_YX5200_SERIAL_PIN_V6 7  //This is the arduino software serial TX pin
#define TANK_MOTOR_A_PWM_PIN_V6 5
#define TANK_MOTOR_A_IN1_PIN_V6 A0
#define TANK_MOTOR_A_IN2_PIN_V6 10
#define TANK_MOTOR_B_PWM_PIN_V6 6
#define TANK_MOTOR_B_IN1_PIN_V6 A2
#define TANK_MOTOR_B_IN2_PIN_V6 A1
#define TANK_RF24_CSN_PIN_V6 A3
#define TANK_RF24_CE_PIN_V6 21		//use the A7(D21) as dummy pin
#define TANK_IR_RECEIVER_PIN_V6 2
#define TANK_IR_SENDER_PIN_V6 3
#define TANK_WS2812_LED_PIN_V6 4		//ws2812 rgb led
#define TANK_FRONT_LED_PIN_V6 8		//A 1w led (visible light)
#define TANK_SERVO1_PIN_V6 9			//tico servo
//====================== V7 board ========================
#define TANK_YX5200_SERIAL_PIN_V7 7  //This is the arduino software serial TX pin
#define TANK_MOTOR_A_PWM_PIN_V7 5
#define TANK_MOTOR_A_IN1_PIN_V7 A0
#define TANK_MOTOR_A_IN2_PIN_V7 8
#define TANK_MOTOR_B_PWM_PIN_V7 6
#define TANK_MOTOR_B_IN1_PIN_V7 A2
#define TANK_MOTOR_B_IN2_PIN_V7 A1
#define TANK_RF24_CSN_PIN_V7 A3
#define TANK_RF24_CE_PIN_V7 21		//use the A7(D21) as dummy pin
#define TANK_IR_RECEIVER_PIN_V7 2
#define TANK_IR_SENDER_PIN_V7 3
#define TANK_WS2812_LED_PIN_V7 4		//ws2812 rgb led
#define TANK_FRONT_LED_PIN_V7 10		//A 1w led (visible light)
#define TANK_SERVO1_PIN_V7 9			//tico servo
//====================== board end ========================


#pragma endregion Tank Pin assignment

#pragma region Tank Development Board mods
/*
serialNumber = 0: Normal version
serialNumber = 1: Small and slower moter, change max speed
*/
#pragma endregion Tank Development Board mods

#pragma region Tank Game play
#define DEFAULT_RELOADING_PERIOD 1500
#define DEFAULT_BLOOD 50
#define DEFAULT_BLOOD_LOW_THRESHOLD 20
#define DEFAULT_BLOOD_CRITICAL_THRESHOLD 10
#define DEFAULT_LIFE 3
#define NORMAL_SHOT_DAMAGE 10
#define BLOOD_LOW_FLASHING_PERIOD 600
#define BLOOD_CRITICAL_FLASHING_PERIOD 200
#define DEFAULT_INVINCIBLE_PERIOD_AFTER_HIT 2000
#define DEFAULT_RELOADING_PERIOD_DURING_INVINCIBLE 500
#define DEFAULT_RELOADING_PERIOD_BLOOD_CRITICAL 100
#define MIN_HEAD_COLOR_INDEX 1
#define MIN_EYES_COLOR_INDEX 1
#define MAX_HEAD_COLOR_INDEX 5 //max 15
#define MAX_EYES_COLOR_INDEX 5 //max 255

typedef enum BLOOD_STATE
{
	BLOOD_STATE_NULL,
	BLOOD_STATE_NORMAL,
	BLOOD_STATE_LOW,
	BLOOD_STATE_CRITICAL,
};

#pragma endregion Tank Game play

#pragma region Tank Mp3
#define BAUD_9600_TX_DELAY 413	//This value is obtained inside of SoftwareSwrial.begin() method.
#define MP3_DEFAULT_VOLUME 20
#define PLAYLIST_MAX_SIZE 8
#define DEFAULT_PLAYBACK_TIME 1500
#define MP3_SF_FIRING_SOUND 1
#define MP3_SF_DEAD 2
#define MP3_SF_REVIVE 3
#define MP3_SF_HIT 4
#define MP3_SF_MAJOR_HIT 5
#define MP3_SF_DEAD_HIT_RANDOM_1 6		//30%
#define MP3_SF_DEAD_HIT_RANDOM_2 7		//10%
#define MP3_SF_DEAD_HIT_RANDOM_3 8		//10%
#define MP3_SF_DEAD_HIT_RANDOM_4 9		//10%
#define MP3_SF_DEAD_HIT_RANDOM_5 10		//10%
#define MP3_SF_DEAD_HIT_RANDOM_6 11		//10%
#define MP3_SF_DEAD_HIT_RANDOM_7 12		//10%
#define MP3_SF_DEAD_HIT_RANDOM_8 13		//10%
#define MP3_SF_OFFLINE 14
#define MP3_SF_ONLINE 15
#define MP3_SF_RECONNECTING 16
#define MP3_SF_INITIALIZATION_COMPLETED 17
#define MP3_SF_GAME_START 18
#define MP3_SF_SINGLE_PVP 19
#define MP3_SF_TEAM_MATCH 20
#define MP3_SF_ADD_TO_BLUE_TEAM 21
#define MP3_SF_ADD_TO_RED_TEAM 22
#define MP3_SF_ADD_TO_PURPLE_TEAM 23
#define MP3_SF_ADD_TO_YELLOW_TEAM 24
#define MP3_SF_ADD_TO_GREEN_TEAM 25
#define RF24_CONNECTION_VOICE_DEFAULT_STATUS false;
#pragma endregion Tank Mp3

#pragma region Tank Servo
#define SERVO1_RETRACTED_POSITION 800
#define SERVO1_HALF_EXTENDED_POSITION 1600
#define SERVO1_FULL_EXTENDED_POSITION 2050
#define SERVO1_FULL_OPERATION_OFF_PERIOD 3000
#define SERVO1_DIRECT_DRIVE_UPDATE_PERIOD 20
#pragma endregion Tank Servo

#pragma region Tank WS2812 led
#define TANK_WS2812_LED_COUNT 5	
#pragma endregion WS2812 led

#pragma region Tank Front LED
#define FRONT_LED_PWM_OFF_PERIOD 10
#pragma endregion Tank Front LED

#pragma region Tank Communication and Control
#define BUTTON_SELECT 0x02
#define BUTTON_START 0x01
#define BUTTON_B 0x80
#define BUTTON_A 0x20
#define JX_CENTER 128
#define JX_LOW_THRESHOLD 40
#define JX_HIGH_THRESHOLD 110
#define JY_CENTER 128
#define JY_LOW_THRESHOLD 40
#define JY_HIGH_THRESHOLD 110
#define DEFAULT_MOTOR_PWM_LOW_THRESHOLD 60
#define DEFAULT_MAX_FORWARD_SPEED 255
#define DEFAULT_MAX_BACKWARD_SPEED 255
#define DEFAULT_MAX_TURN_SPEED 80
#define DEFAULT_DIRECT_MAX_SPEED 255
#define SIGNAL_LOSS_MOTOR_OFF_PERIOD 200
#define DEFAULT_MOTOR_FILTER_SHIFT 2
#define IR_COMMAND_NORMAL_FIRE 0x28

enum PACKET_CMD_BYTE
{
	CMD_NULL = 0,
	CMD_BASIC_CONTROL_V1,
	CMD_MAX_SIZE,
};

enum ACK_BYTE
{
	ACK_NULL = 0,
	ACK_CMD_V1_OK,
	ACK_MAX_SIZE,
};

typedef enum TANK_STATE
{
	TANK_STATE_NULL,
	TANK_STATE_PLAYING,
	TANK_STATE_DEAD,
	TANK_STATE_GAME_SELECTION,
};

typedef enum GAME_MODE
{
	GAME_MODE_NULL,
	GAME_MODE_SINGLE,
	GAME_MODE_TEAM,
};

/*IR Packet description
Byte 1:
Bit 7:4 Game mode
Bit 3:0 Sender ID (Head color index)
Byte 2:
Reserved for armor and base
Byte 3:
Bit 7:0 Command Value

*/

/*RF24 Packet description
Byte 0: PACKET_CMD_BYTE

================================================
PACKET_CMD_BYTE = CMD_BASIC_CONTROL_V1
Byte 1: Joystick X value 0~255. 127=center
Byte 2: Joystick Y value 0~255. 127=center
Byte 3: Four buttons state: select, start, B, A
================================================

*/

#pragma endregion Tank Communication and Control

class FuzzyTank
{
public:

	/**
	* Arduino Constructor
	* Creates a new instance of this class.
	*/
	FuzzyTank();

	/**
	* Begin the operation of the tank board.
	*
	* Call this in setup(), before calling any other methods.
	* @code
	* tank.begin()
	* @endcode
	*/
	void begin();

	/*Initialization with user assigned board version*/
	void begin(uint8_t boardVersion);

	/**
	* Play a numbered file in the SD card \mp3 folder.
	*
	* @code
	* tank.playMp3(1);//plays the 001.mp3 file in the \mp3 folder.
	* @code
	*
	* The file can have descriptions after the numbers, for example:
	* 001_game_start.mp3 can be played by tank.playMp3(1)
	*
	*/
	void playMp3(uint16_t fileNumber);

	void setVolume(uint8_t vol);

	void update();

	void setEyesColor(uint8_t* color);

	void setEyesColor(colorRGB color);

	void setHeadColor(uint8_t* color);

	void setHeadColor(colorRGB color);

	void updateColor();

	void setMaximumSpeed(uint8_t speed);

	void setMaximumForwardSpeed(uint8_t speed);

	void setMaximumBackwardSpeed(uint8_t speed);

	void setMotorSmoothness(uint8_t value);

	void motorDebug(uint8_t x, uint8_t y);

	void setMaximumTurnSpeed(uint8_t speed);

	void setMotorType(MOTOR_TYPE motor);

	void setGearsetType(GEARSET_TYPE gearset);

	void setRF24ConnectionVoicePrompt(bool setting);

#pragma region Game play
	void setBlood(uint8_t blood);

	void setLife(uint8_t life);

	void canBeHitAfterGameover(bool value);

#pragma endregion Game play
private:

#pragma region Setup and pairing
	void setDevelopmentBoardDifference();
	void autoPairing();
#pragma endregion Setup and pairing

#pragma region Pin assignment
	void setPins();
	uint8_t boardVersion;
	const uint8_t validBoardVersions[TANK_SUPPORTED_BOARD_TYPES] = { 2,3,4,5,6,7 };
	uint8_t yx5200SerialPin;
	uint8_t motorAPwmPin;
	uint8_t motorAIn1Pin;
	uint8_t motorAIn2Pin;
	uint8_t motorBPwmPin;
	uint8_t motorBIn1Pin;
	uint8_t motorBIn2Pin;
	uint8_t RF24CsnPin;
	uint8_t RF24CePin;
	uint8_t IRReceiverPin;
	uint8_t IRSenderPin;
	uint8_t WS2812LedPin;
	uint8_t frontLedPin;
	uint8_t servo1Pin;
#pragma endregion Pin assignment

#pragma region MP3
	uint8_t mp3TxBuffer[10];
	void mp3Begin();
	void writeChecksum();
	void sendBufferToMp3();

	void playListPush(uint16_t value);
	uint16_t playListPop();
	uint8_t playListSize();
	bool playListIsEmpty();
	void updateMp3();
	uint8_t items = 0;
	uint8_t front = 0;
	uint8_t rear = 0;
	uint16_t playListData[PLAYLIST_MAX_SIZE];
	uint32_t playbackTime = DEFAULT_PLAYBACK_TIME;
	uint32_t playbackTimer;
	bool rf24ConnectionVoiceStatus = RF24_CONNECTION_VOICE_DEFAULT_STATUS;
#pragma endregion MP3

#pragma region SoftwareSerialSendOnly
	void softSerialBegin();
	inline void tunedDelay(uint16_t delay);
	void serialWrite(uint8_t b);
	uint16_t _tx_delay = BAUD_9600_TX_DELAY;
	uint8_t _transmitBitMask;
	volatile uint8_t *_transmitPortRegister;
	bool _inverse_logic = false;
#pragma endregion SoftwareSerialSendOnly

#pragma region MOTOR Control
	void motorBegin();
	void commandMotor(int8_t motorLeft, int8_t motorRight);
	void commandMotorDirect(int8_t motorLeft, int8_t motorRight);
	int32_t motorLeftFilterRegister;
	int32_t motorRightFilterRegister;
	int8_t motorLeftOutput;
	int8_t motorRightOutput;
	uint8_t motorFilterShift;
	uint8_t motorLeftPWM;
	uint8_t motorRightPWM;
	uint8_t motorPWMLowThreshold;
	MOTOR_TYPE motorType;
	GEARSET_TYPE gearsetType;
	int8_t motorOutputDirection = 1;

#ifdef DEBUG_MOTOR_MOVEMENT
	void updateDebugMotor();
	bool debugMotorState;
	int8_t debugMotorSpeed;
#endif //DEBUG_MOTOR_MOVEMENT
#pragma endregion MOTOR Control

#pragma region WS2812 led
	colorRGB color;
	colorRGB eyesColor;
	colorRGB headColor;
	/* fixed colors */

	uint8_t fixColor[8][3] = {
		{ 0,0,0 },//C_OFF
		{ 0,174,219 },//C_BLUE
		{ 212,18,67 },//C_RED
		{ 162,0,255 },//C_PURPLE
		{ 232,167,20 },//C_YELLOW
					   //{ 244,120,53 },//C_ORANGE
		{ 142,233,39 },//C_GREEN
					   //{ 142,193,39 },//C_GREEN
		{ 224,202,81 },//C_WARM_WHITE
		{ 255,255,255 },//C_BRIGHT_WHITE
	};

	enum color {
		C_OFF,
		C_BLUE,
		C_RED,
		C_PURPLE,
		C_YELLOW,
		C_GREEN,
		C_WARM_WHITE,
		C_BRIGHT_WHITE,
	};

	/*	uint8_t off[3] = { 0,0,0 };
	uint8_t warmWhite[3] = { 224,202,81 };
	uint8_t brightWhite[3] = { 255,255,255 };
	uint8_t blue[3] = { 0,174,219 };
	uint8_t purple[3] = { 162,0,255 };
	uint8_t orange[3] = { 244,120,53 };
	uint8_t red[3] = { 212,18,67 };
	uint8_t green[3] = { 212,18,67 };
	uint8_t royalBlue[3] = { 65,105,225 };*/
	void ws2812Begin();
	uint8_t pixels[TANK_WS2812_LED_COUNT * 3];	//the array to hold ws2812 rgb color values, three bytes for each pixel
	void setColor(uint8_t pixelIndex, colorRGB color);
	void setColor(uint8_t pixelIndex, uint8_t* color);
	uint8_t pinMask;
	const volatile uint8_t *ws2812_port;
	volatile uint8_t *ws2812_port_reg;
	void ws2812_sendarray_mask(uint8_t *data, uint16_t datlen, uint8_t maskhi, uint8_t *port, uint8_t *portreg);
	void updateRGBLed();
	bool blinkEyes = true;
	uint32_t blinkEyesTimer;
	uint32_t blinkEyesTimeDuration;
	bool eyesLedIsOn;
#pragma endregion WS2812

#pragma region Front LED
	void frontLedBegin();
	void frontLedOn(uint32_t time);
	uint32_t frontLedOnTimer;
	uint32_t frontLedDuration;
	void updateFrontLed();
	bool frontLedTimerEnabled = false;
	bool frontLedPWMEnabled = false;
	bool frontLedPWMState = false;
	uint32_t frontLedPwmTimer;
	void setFrontLedDimOn();
	void setFrontLedDimOff();
#pragma endregion Front LED

#pragma region Servo
	void servoBegin();
	void enableServo1();
	void disableServo1();
	void updateServo();
	void setServo1(uint16_t position);
	void setServo1AndOff(uint16_t position, uint32_t period); //set the servo to a position, start a timer, and turn the power off when time's up
	bool servo1IsEnabled = false;
	bool servo1OffTimerIsEnabled = false;
	uint32_t servo1OffTimer = 0;;
	uint32_t servo1OffPeroid = 0;
	uint32_t servo1DirectDriveTimer;
	uint16_t servo1DelayLoop2Count;
	void delayWithServoUpdate(uint16_t duration);
#pragma endregion Servo

#pragma region RF24
	/**
	* @name Primary public interface
	*
	*  These are the main methods you need to operate the chip
	*/
	/**@{*/

	/**
	* Arduino Constructor
	*
	* Creates a new instance of this driver.  Before using, you create an instance
	* and send in the unique pins that this chip is connected to.
	*
	* @param _cepin The pin attached to Chip Enable on the RF module
	* @param _cspin The pin attached to Chip Select
	*/
	FuzzyTank(uint16_t _cepin, uint16_t _cspin);
#if defined (RF24_LINUX)

	/**
	* Optional Linux Constructor
	*
	* Creates a new instance of this driver.  Before using, you create an instance
	* and send in the unique pins that this chip is connected to.
	*
	* @param _cepin The pin attached to Chip Enable on the RF module
	* @param _cspin The pin attached to Chip Select
	* @param spispeed For RPi, the SPI speed in MHZ ie: BCM2835_SPI_SPEED_8MHZ
	*/

	FuzzyTank(uint16_t _cepin, uint16_t _cspin, uint32_t spispeed);
#	endif

#if defined (RF24_LINUX)
	virtual ~RF24() {};
#endif

	/**
	* Begin operation of the chip
	*
	* Call this in setup(), before calling any other methods.
	* @code radio.begin() @endcode
	*/
	bool RF24Begin(void);

	/**
	* Checks if the chip is connected to the SPI bus
	*/
	bool isChipConnected();

	/**
	* Start listening on the pipes opened for reading.
	*
	* 1. Be sure to call openReadingPipe() first.
	* 2. Do not call write() while in this mode, without first calling stopListening().
	* 3. Call available() to check for incoming traffic, and read() to get it.
	*
	* @code
	* Open reading pipe 1 using address CCCECCCECC
	*
	* byte address[] = { 0xCC,0xCE,0xCC,0xCE,0xCC };
	* radio.openReadingPipe(1,address);
	* radio.startListening();
	* @endcode
	*/
	void startListening(void);

	/**
	* Stop listening for incoming messages, and switch to transmit mode.
	*
	* Do this before calling write().
	* @code
	* radio.stopListening();
	* radio.write(&data,sizeof(data));
	* @endcode
	*/
	void stopListening(void);

	/**
	* Check whether there are bytes available to be read
	* @code
	* if(radio.available()){
	*   radio.read(&data,sizeof(data));
	* }
	* @endcode
	* @return True if there is a payload available, false if none is
	*/
	bool available(void);

	/**
	* Read the available payload
	*
	* The size of data read is the fixed payload size, see getPayloadSize()
	*
	* @note I specifically chose 'void*' as a data type to make it easier
	* for beginners to use.  No casting needed.
	*
	* @note No longer boolean. Use available to determine if packets are
	* available. Interrupt flags are now cleared during reads instead of
	* when calling available().
	*
	* @param buf Pointer to a buffer where the data should be written
	* @param len Maximum number of bytes to read into the buffer
	*
	* @code
	* if(radio.available()){
	*   radio.read(&data,sizeof(data));
	* }
	* @endcode
	* @return No return value. Use available().
	*/
	void read(void* buf, uint8_t len);

	/**
	* Be sure to call openWritingPipe() first to set the destination
	* of where to write to.
	*
	* This blocks until the message is successfully acknowledged by
	* the receiver or the timeout/retransmit maxima are reached.  In
	* the current configuration, the max delay here is 60-70ms.
	*
	* The maximum size of data written is the fixed payload size, see
	* getPayloadSize().  However, you can write less, and the remainder
	* will just be filled with zeroes.
	*
	* TX/RX/RT interrupt flags will be cleared every time write is called
	*
	* @param buf Pointer to the data to be sent
	* @param len Number of bytes to be sent
	*
	* @code
	* radio.stopListening();
	* radio.write(&data,sizeof(data));
	* @endcode
	* @return True if the payload was delivered successfully false if not
	*/
	bool write(const void* buf, uint8_t len);

	/**
	* New: Open a pipe for writing via byte array. Old addressing format retained
	* for compatibility.
	*
	* Only one writing pipe can be open at once, but you can change the address
	* you'll write to. Call stopListening() first.
	*
	* Addresses are assigned via a byte array, default is 5 byte address length
	s   *
	* @code
	*   uint8_t addresses[][6] = {"1Node","2Node"};
	*   radio.openWritingPipe(addresses[0]);
	* @endcode
	* @code
	*  uint8_t address[] = { 0xCC,0xCE,0xCC,0xCE,0xCC };
	*  radio.openWritingPipe(address);
	*  address[0] = 0x33;
	*  radio.openReadingPipe(1,address);
	* @endcode
	* @see setAddressWidth
	*
	* @param address The address of the pipe to open. Coordinate these pipe
	* addresses amongst nodes on the network.
	*/

	void openWritingPipe(const uint8_t *address);

	/**
	* Open a pipe for reading
	*
	* Up to 6 pipes can be open for reading at once.  Open all the required
	* reading pipes, and then call startListening().
	*
	* @see openWritingPipe
	* @see setAddressWidth
	*
	* @note Pipes 0 and 1 will store a full 5-byte address. Pipes 2-5 will technically
	* only store a single byte, borrowing up to 4 additional bytes from pipe #1 per the
	* assigned address width.
	* @warning Pipes 1-5 should share the same address, except the first byte.
	* Only the first byte in the array should be unique, e.g.
	* @code
	*   uint8_t addresses[][6] = {"1Node","2Node"};
	*   openReadingPipe(1,addresses[0]);
	*   openReadingPipe(2,addresses[1]);
	* @endcode
	*
	* @warning Pipe 0 is also used by the writing pipe.  So if you open
	* pipe 0 for reading, and then startListening(), it will overwrite the
	* writing pipe.  Ergo, do an openWritingPipe() again before write().
	*
	* @param number Which pipe# to open, 0-5.
	* @param address The 24, 32 or 40 bit address of the pipe to open.
	*/

	void openReadingPipe(uint8_t number, const uint8_t *address);

	/**@}*/
	/**
	* @name Advanced Operation
	*
	*  Methods you can use to drive the chip in more advanced ways
	*/
	/**@{*/

	/**
	* Print a giant block of debugging information to stdout
	*
	* @warning Does nothing if stdout is not defined.  See fdevopen in stdio.h
	* The printf.h file is included with the library for Arduino.
	* @code
	* #include <printf.h>
	* setup(){
	*  Serial.begin(115200);
	*  printf_begin();
	*  ...
	* }
	* @endcode
	*/
	void printDetails(void);

	/**
	* Test whether there are bytes available to be read in the
	* FIFO buffers.
	*
	* @param[out] pipe_num Which pipe has the payload available
	*
	* @code
	* uint8_t pipeNum;
	* if(radio.available(&pipeNum)){
	*   radio.read(&data,sizeof(data));
	*   Serial.print("Got data on pipe");
	*   Serial.println(pipeNum);
	* }
	* @endcode
	* @return True if there is a payload available, false if none is
	*/
	bool available(uint8_t* pipe_num);

	/**
	* Check if the radio needs to be read. Can be used to prevent data loss
	* @return True if all three 32-byte radio buffers are full
	*/
	bool rxFifoFull();

	/**
	* Enter low-power mode
	*
	* To return to normal power mode, call powerUp().
	*
	* @note After calling startListening(), a basic radio will consume about 13.5mA
	* at max PA level.
	* During active transmission, the radio will consume about 11.5mA, but this will
	* be reduced to 26uA (.026mA) between sending.
	* In full powerDown mode, the radio will consume approximately 900nA (.0009mA)
	*
	* @code
	* radio.powerDown();
	* avr_enter_sleep_mode(); // Custom function to sleep the device
	* radio.powerUp();
	* @endcode
	*/
	void powerDown(void);

	/**
	* Leave low-power mode - required for normal radio operation after calling powerDown()
	*
	* To return to low power mode, call powerDown().
	* @note This will take up to 5ms for maximum compatibility
	*/
	void powerUp(void);

	/**
	* Write for single NOACK writes. Optionally disables acknowledgements/autoretries for a single write.
	*
	* @note enableDynamicAck() must be called to enable this feature
	*
	* Can be used with enableAckPayload() to request a response
	* @see enableDynamicAck()
	* @see setAutoAck()
	* @see write()
	*
	* @param buf Pointer to the data to be sent
	* @param len Number of bytes to be sent
	* @param multicast Request ACK (0), NOACK (1)
	*/
	bool write(const void* buf, uint8_t len, const bool multicast);

	/**
	* This will not block until the 3 FIFO buffers are filled with data.
	* Once the FIFOs are full, writeFast will simply wait for success or
	* timeout, and return 1 or 0 respectively. From a user perspective, just
	* keep trying to send the same data. The library will keep auto retrying
	* the current payload using the built in functionality.
	* @warning It is important to never keep the nRF24L01 in TX mode and FIFO full for more than 4ms at a time. If the auto
	* retransmit is enabled, the nRF24L01 is never in TX mode long enough to disobey this rule. Allow the FIFO
	* to clear by issuing txStandBy() or ensure appropriate time between transmissions.
	*
	* @code
	* Example (Partial blocking):
	*
	*			radio.writeFast(&buf,32);  // Writes 1 payload to the buffers
	*			txStandBy();     		   // Returns 0 if failed. 1 if success. Blocks only until MAX_RT timeout or success. Data flushed on fail.
	*
	*			radio.writeFast(&buf,32);  // Writes 1 payload to the buffers
	*			txStandBy(1000);		   // Using extended timeouts, returns 1 if success. Retries failed payloads for 1 seconds before returning 0.
	* @endcode
	*
	* @see txStandBy()
	* @see write()
	* @see writeBlocking()
	*
	* @param buf Pointer to the data to be sent
	* @param len Number of bytes to be sent
	* @return True if the payload was delivered successfully false if not
	*/
	bool writeFast(const void* buf, uint8_t len);

	/**
	* WriteFast for single NOACK writes. Disables acknowledgements/autoretries for a single write.
	*
	* @note enableDynamicAck() must be called to enable this feature
	* @see enableDynamicAck()
	* @see setAutoAck()
	*
	* @param buf Pointer to the data to be sent
	* @param len Number of bytes to be sent
	* @param multicast Request ACK (0) or NOACK (1)
	*/
	bool writeFast(const void* buf, uint8_t len, const bool multicast);

	/**
	* This function extends the auto-retry mechanism to any specified duration.
	* It will not block until the 3 FIFO buffers are filled with data.
	* If so the library will auto retry until a new payload is written
	* or the user specified timeout period is reached.
	* @warning It is important to never keep the nRF24L01 in TX mode and FIFO full for more than 4ms at a time. If the auto
	* retransmit is enabled, the nRF24L01 is never in TX mode long enough to disobey this rule. Allow the FIFO
	* to clear by issuing txStandBy() or ensure appropriate time between transmissions.
	*
	* @code
	* Example (Full blocking):
	*
	*			radio.writeBlocking(&buf,32,1000); //Wait up to 1 second to write 1 payload to the buffers
	*			txStandBy(1000);     			   //Wait up to 1 second for the payload to send. Return 1 if ok, 0 if failed.
	*					  				   		   //Blocks only until user timeout or success. Data flushed on fail.
	* @endcode
	* @note If used from within an interrupt, the interrupt should be disabled until completion, and sei(); called to enable millis().
	* @see txStandBy()
	* @see write()
	* @see writeFast()
	*
	* @param buf Pointer to the data to be sent
	* @param len Number of bytes to be sent
	* @param timeout User defined timeout in milliseconds.
	* @return True if the payload was loaded into the buffer successfully false if not
	*/
	bool writeBlocking(const void* buf, uint8_t len, uint32_t timeout);

	/**
	* This function should be called as soon as transmission is finished to
	* drop the radio back to STANDBY-I mode. If not issued, the radio will
	* remain in STANDBY-II mode which, per the data sheet, is not a recommended
	* operating mode.
	*
	* @note When transmitting data in rapid succession, it is still recommended by
	* the manufacturer to drop the radio out of TX or STANDBY-II mode if there is
	* time enough between sends for the FIFOs to empty. This is not required if auto-ack
	* is enabled.
	*
	* Relies on built-in auto retry functionality.
	*
	* @code
	* Example (Partial blocking):
	*
	*			radio.writeFast(&buf,32);
	*			radio.writeFast(&buf,32);
	*			radio.writeFast(&buf,32);  //Fills the FIFO buffers up
	*			bool ok = txStandBy();     //Returns 0 if failed. 1 if success.
	*					  				   //Blocks only until MAX_RT timeout or success. Data flushed on fail.
	* @endcode
	* @see txStandBy(unsigned long timeout)
	* @return True if transmission is successful
	*
	*/
	bool txStandBy();

	/**
	* This function allows extended blocking and auto-retries per a user defined timeout
	* @code
	*	Fully Blocking Example:
	*
	*			radio.writeFast(&buf,32);
	*			radio.writeFast(&buf,32);
	*			radio.writeFast(&buf,32);   //Fills the FIFO buffers up
	*			bool ok = txStandBy(1000);  //Returns 0 if failed after 1 second of retries. 1 if success.
	*					  				    //Blocks only until user defined timeout or success. Data flushed on fail.
	* @endcode
	* @note If used from within an interrupt, the interrupt should be disabled until completion, and sei(); called to enable millis().
	* @param timeout Number of milliseconds to retry failed payloads
	* @return True if transmission is successful
	*
	*/
	bool txStandBy(uint32_t timeout, bool startTx = 0);

	/**
	* Write an ack payload for the specified pipe
	*
	* The next time a message is received on @p pipe, the data in @p buf will
	* be sent back in the acknowledgement.
	* @see enableAckPayload()
	* @see enableDynamicPayloads()
	* @warning Only three of these can be pending at any time as there are only 3 FIFO buffers.<br> Dynamic payloads must be enabled.
	* @note Ack payloads are handled automatically by the radio chip when a payload is received. Users should generally
	* write an ack payload as soon as startListening() is called, so one is available when a regular payload is received.
	* @note Ack payloads are dynamic payloads. This only works on pipes 0&1 by default. Call
	* enableDynamicPayloads() to enable on all pipes.
	*
	* @param pipe Which pipe# (typically 1-5) will get this response.
	* @param buf Pointer to data that is sent
	* @param len Length of the data to send, up to 32 bytes max.  Not affected
	* by the static payload set by setPayloadSize().
	*/
	void writeAckPayload(uint8_t pipe, const void* buf, uint8_t len);

	/**
	* Determine if an ack payload was received in the most recent call to
	* write(). The regular available() can also be used.
	*
	* Call read() to retrieve the ack payload.
	*
	* @return True if an ack payload is available.
	*/
	bool isAckPayloadAvailable(void);

	/**
	* Call this when you get an interrupt to find out why
	*
	* Tells you what caused the interrupt, and clears the state of
	* interrupts.
	*
	* @param[out] tx_ok The send was successful (TX_DS)
	* @param[out] tx_fail The send failed, too many retries (MAX_RT)
	* @param[out] rx_ready There is a message waiting to be read (RX_DS)
	*/
	void whatHappened(bool& tx_ok, bool& tx_fail, bool& rx_ready);

	/**
	* Non-blocking write to the open writing pipe used for buffered writes
	*
	* @note Optimization: This function now leaves the CE pin high, so the radio
	* will remain in TX or STANDBY-II Mode until a txStandBy() command is issued. Can be used as an alternative to startWrite()
	* if writing multiple payloads at once.
	* @warning It is important to never keep the nRF24L01 in TX mode with FIFO full for more than 4ms at a time. If the auto
	* retransmit/autoAck is enabled, the nRF24L01 is never in TX mode long enough to disobey this rule. Allow the FIFO
	* to clear by issuing txStandBy() or ensure appropriate time between transmissions.
	*
	* @see write()
	* @see writeFast()
	* @see startWrite()
	* @see writeBlocking()
	*
	* For single noAck writes see:
	* @see enableDynamicAck()
	* @see setAutoAck()
	*
	* @param buf Pointer to the data to be sent
	* @param len Number of bytes to be sent
	* @param multicast Request ACK (0) or NOACK (1)
	* @return True if the payload was delivered successfully false if not
	*/
	void startFastWrite(const void* buf, uint8_t len, const bool multicast, bool startTx = 1);

	/**
	* Non-blocking write to the open writing pipe
	*
	* Just like write(), but it returns immediately. To find out what happened
	* to the send, catch the IRQ and then call whatHappened().
	*
	* @see write()
	* @see writeFast()
	* @see startFastWrite()
	* @see whatHappened()
	*
	* For single noAck writes see:
	* @see enableDynamicAck()
	* @see setAutoAck()
	*
	* @param buf Pointer to the data to be sent
	* @param len Number of bytes to be sent
	* @param multicast Request ACK (0) or NOACK (1)
	*
	*/
	void startWrite(const void* buf, uint8_t len, const bool multicast);

	/**
	* This function is mainly used internally to take advantage of the auto payload
	* re-use functionality of the chip, but can be beneficial to users as well.
	*
	* The function will instruct the radio to re-use the data in the FIFO buffers,
	* and instructs the radio to re-send once the timeout limit has been reached.
	* Used by writeFast and writeBlocking to initiate retries when a TX failure
	* occurs. Retries are automatically initiated except with the standard write().
	* This way, data is not flushed from the buffer until switching between modes.
	*
	* @note This is to be used AFTER auto-retry fails if wanting to resend
	* using the built-in payload reuse features.
	* After issuing reUseTX(), it will keep reending the same payload forever or until
	* a payload is written to the FIFO, or a flush_tx command is given.
	*/
	void reUseTX();

	/**
	* Empty the transmit buffer. This is generally not required in standard operation.
	* May be required in specific cases after stopListening() , if operating at 250KBPS data rate.
	*
	* @return Current value of status register
	*/
	uint8_t flush_tx(void);

	/**
	* Test whether there was a carrier on the line for the
	* previous listening period.
	*
	* Useful to check for interference on the current channel.
	*
	* @return true if was carrier, false if not
	*/
	bool testCarrier(void);

	/**
	* Test whether a signal (carrier or otherwise) greater than
	* or equal to -64dBm is present on the channel. Valid only
	* on nRF24L01P (+) hardware. On nRF24L01, use testCarrier().
	*
	* Useful to check for interference on the current channel and
	* channel hopping strategies.
	*
	* @code
	* bool goodSignal = radio.testRPD();
	* if(radio.available()){
	*    Serial.println(goodSignal ? "Strong signal > 64dBm" : "Weak signal < 64dBm" );
	*    radio.read(0,0);
	* }
	* @endcode
	* @return true if signal => -64dBm, false if not
	*/
	bool testRPD(void);

	/**
	* Test whether this is a real radio, or a mock shim for
	* debugging.  Setting either pin to 0xff is the way to
	* indicate that this is not a real radio.
	*
	* @return true if this is a legitimate radio
	*/
	bool isValid() { return ce_pin != 0xff && csn_pin != 0xff; }

	/**
	* Close a pipe after it has been previously opened.
	* Can be safely called without having previously opened a pipe.
	* @param pipe Which pipe # to close, 0-5.
	*/
	void closeReadingPipe(uint8_t pipe);

	/**
	* Enable error detection by un-commenting #define FAILURE_HANDLING in RF24_config.h
	* If a failure has been detected, it usually indicates a hardware issue. By default the library
	* will cease operation when a failure is detected.
	* This should allow advanced users to detect and resolve intermittent hardware issues.
	*
	* In most cases, the radio must be re-enabled via radio.begin(); and the appropriate settings
	* applied after a failure occurs, if wanting to re-enable the device immediately.
	*
	* Usage: (Failure handling must be enabled per above)
	*  @code
	*  if(radio.failureDetected){
	*    radio.begin();                       // Attempt to re-configure the radio with defaults
	*    radio.failureDetected = 0;           // Reset the detection value
	*	radio.openWritingPipe(addresses[1]); // Re-configure pipe addresses
	*    radio.openReadingPipe(1,addresses[0]);
	*    report_failure();                    // Blink leds, send a message, etc. to indicate failure
	*  }
	* @endcode
	*/
	//#if defined (FAILURE_HANDLING)
	bool failureDetected;
	//#endif

	/**@}*/

	/**@}*/
	/**
	* @name Optional Configurators
	*
	*  Methods you can use to get or set the configuration of the chip.
	*  None are required.  Calling begin() sets up a reasonable set of
	*  defaults.
	*/
	/**@{*/

	/**
	* Set the address width from 3 to 5 bytes (24, 32 or 40 bit)
	*
	* @param a_width The address width to use: 3,4 or 5
	*/

	void setAddressWidth(uint8_t a_width);

	/**
	* Set the number and delay of retries upon failed submit
	*
	* @param delay How long to wait between each retry, in multiples of 250us,
	* max is 15.  0 means 250us, 15 means 4000us.
	* @param count How many retries before giving up, max 15
	*/
	void setRetries(uint8_t delay, uint8_t count);

	/**
	* Set RF communication channel
	*
	* @param channel Which RF channel to communicate on, 0-125
	*/
	void setChannel(uint8_t channel);

	/**
	* Get RF communication channel
	*
	* @return The currently configured RF Channel
	*/
	uint8_t getChannel(void);

	/**
	* Set Static Payload Size
	*
	* This implementation uses a pre-stablished fixed payload size for all
	* transmissions.  If this method is never called, the driver will always
	* transmit the maximum payload size (32 bytes), no matter how much
	* was sent to write().
	*
	* @todo Implement variable-sized payloads feature
	*
	* @param size The number of bytes in the payload
	*/
	void setPayloadSize(uint8_t size);

	/**
	* Get Static Payload Size
	*
	* @see setPayloadSize()
	*
	* @return The number of bytes in the payload
	*/
	uint8_t getPayloadSize(void);

	/**
	* Get Dynamic Payload Size
	*
	* For dynamic payloads, this pulls the size of the payload off
	* the chip
	*
	* @note Corrupt packets are now detected and flushed per the
	* manufacturer.
	* @code
	* if(radio.available()){
	*   if(radio.getDynamicPayloadSize() < 1){
	*     // Corrupt payload has been flushed
	*     return;
	*   }
	*   radio.read(&data,sizeof(data));
	* }
	* @endcode
	*
	* @return Payload length of last-received dynamic payload
	*/
	uint8_t getDynamicPayloadSize(void);

	/**
	* Enable custom payloads on the acknowledge packets
	*
	* Ack payloads are a handy way to return data back to senders without
	* manually changing the radio modes on both units.
	*
	* @note Ack payloads are dynamic payloads. This only works on pipes 0&1 by default. Call
	* enableDynamicPayloads() to enable on all pipes.
	*/
	void enableAckPayload(void);

	/**
	* Enable dynamically-sized payloads
	*
	* This way you don't always have to send large packets just to send them
	* once in a while.  This enables dynamic payloads on ALL pipes.
	*
	*/
	void enableDynamicPayloads(void);

	/**
	* Disable dynamically-sized payloads
	*
	* This disables dynamic payloads on ALL pipes. Since Ack Payloads
	* requires Dynamic Payloads, Ack Payloads are also disabled.
	* If dynamic payloads are later re-enabled and ack payloads are desired
	* then enableAckPayload() must be called again as well.
	*
	*/
	void disableDynamicPayloads(void);

	/**
	* Enable dynamic ACKs (single write multicast or unicast) for chosen messages
	*
	* @note To enable full multicast or per-pipe multicast, use setAutoAck()
	*
	* @warning This MUST be called prior to attempting single write NOACK calls
	* @code
	* radio.enableDynamicAck();
	* radio.write(&data,32,1);  // Sends a payload with no acknowledgement requested
	* radio.write(&data,32,0);  // Sends a payload using auto-retry/autoACK
	* @endcode
	*/
	void enableDynamicAck();

	/**
	* Determine whether the hardware is an nRF24L01+ or not.
	*
	* @return true if the hardware is nRF24L01+ (or compatible) and false
	* if its not.
	*/
	bool isPVariant(void);

	/**
	* Enable or disable auto-acknowlede packets
	*
	* This is enabled by default, so it's only needed if you want to turn
	* it off for some reason.
	*
	* @param enable Whether to enable (true) or disable (false) auto-acks
	*/
	void setAutoAck(bool enable);

	/**
	* Enable or disable auto-acknowlede packets on a per pipeline basis.
	*
	* AA is enabled by default, so it's only needed if you want to turn
	* it off/on for some reason on a per pipeline basis.
	*
	* @param pipe Which pipeline to modify
	* @param enable Whether to enable (true) or disable (false) auto-acks
	*/
	void setAutoAck(uint8_t pipe, bool enable);

	/**
	* Set Power Amplifier (PA) level to one of four levels:
	* RF24_PA_MIN, RF24_PA_LOW, RF24_PA_HIGH and RF24_PA_MAX
	*
	* The power levels correspond to the following output levels respectively:
	* NRF24L01: -18dBm, -12dBm,-6dBM, and 0dBm
	*
	* SI24R1: -6dBm, 0dBm, 3dBM, and 7dBm.
	*
	* @param level Desired PA level.
	*/
	void setPALevel(uint8_t level);

	/**
	* Fetches the current PA level.
	*
	* NRF24L01: -18dBm, -12dBm, -6dBm and 0dBm
	* SI24R1:   -6dBm, 0dBm, 3dBm, 7dBm
	*
	* @return Returns values 0 to 3 representing the PA Level.
	*/
	uint8_t getPALevel(void);

	/**
	* Set the transmission data rate
	*
	* @warning setting RF24_250KBPS will fail for non-plus units
	*
	* @param speed RF24_250KBPS for 250kbs, RF24_1MBPS for 1Mbps, or RF24_2MBPS for 2Mbps
	* @return true if the change was successful
	*/
	bool setDataRate(rf24_datarate_e speed);

	/**
	* Fetches the transmission data rate
	*
	* @return Returns the hardware's currently configured datarate. The value
	* is one of 250kbs, RF24_1MBPS for 1Mbps, or RF24_2MBPS, as defined in the
	* rf24_datarate_e enum.
	*/
	rf24_datarate_e getDataRate(void);

	/**
	* Set the CRC length
	* <br>CRC checking cannot be disabled if auto-ack is enabled
	* @param length RF24_CRC_8 for 8-bit or RF24_CRC_16 for 16-bit
	*/
	void setCRCLength(rf24_crclength_e length);

	/**
	* Get the CRC length
	* <br>CRC checking cannot be disabled if auto-ack is enabled
	* @return RF24_CRC_DISABLED if disabled or RF24_CRC_8 for 8-bit or RF24_CRC_16 for 16-bit
	*/
	rf24_crclength_e getCRCLength(void);

	/**
	* Disable CRC validation
	*
	* @warning CRC cannot be disabled if auto-ack/ESB is enabled.
	*/
	void disableCRC(void);

	/**
	* The radio will generate interrupt signals when a transmission is complete,
	* a transmission fails, or a payload is received. This allows users to mask
	* those interrupts to prevent them from generating a signal on the interrupt
	* pin. Interrupts are enabled on the radio chip by default.
	*
	* @code
	* 	Mask all interrupts except the receive interrupt:
	*
	*		radio.maskIRQ(1,1,0);
	* @endcode
	*
	* @param tx_ok  Mask transmission complete interrupts
	* @param tx_fail  Mask transmit failure interrupts
	* @param rx_ready Mask payload received interrupts
	*/
	void maskIRQ(bool tx_ok, bool tx_fail, bool rx_ready);

	/**
	*
	* The driver will delay for this duration when stopListening() is called
	*
	* When responding to payloads, faster devices like ARM(RPi) are much faster than Arduino:
	* 1. Arduino sends data to RPi, switches to RX mode
	* 2. The RPi receives the data, switches to TX mode and sends before the Arduino radio is in RX mode
	* 3. If AutoACK is disabled, this can be set as low as 0. If AA/ESB enabled, set to 100uS minimum on RPi
	*
	* @warning If set to 0, ensure 130uS delay after stopListening() and before any sends
	*/

	uint32_t txDelay;

	/**
	*
	* On all devices but Linux and ATTiny, a small delay is added to the CSN toggling function
	*
	* This is intended to minimise the speed of SPI polling due to radio commands
	*
	* If using interrupts or timed requests, this can be set to 0 Default:5
	*/

	uint32_t csDelay;

	/**@}*/
	/**
	* @name Deprecated
	*
	*  Methods provided for backwards compabibility.
	*/
	/**@{*/


	/**
	* Open a pipe for reading
	* @note For compatibility with old code only, see new function
	*
	* @warning Pipes 1-5 should share the first 32 bits.
	* Only the least significant byte should be unique, e.g.
	* @code
	*   openReadingPipe(1,0xF0F0F0F0AA);
	*   openReadingPipe(2,0xF0F0F0F066);
	* @endcode
	*
	* @warning Pipe 0 is also used by the writing pipe.  So if you open
	* pipe 0 for reading, and then startListening(), it will overwrite the
	* writing pipe.  Ergo, do an openWritingPipe() again before write().
	*
	* @param number Which pipe# to open, 0-5.
	* @param address The 40-bit address of the pipe to open.
	*/
	void openReadingPipe(uint8_t number, uint64_t address);

	/**
	* Open a pipe for writing
	* @note For compatibility with old code only, see new function
	*
	* Addresses are 40-bit hex values, e.g.:
	*
	* @code
	*   openWritingPipe(0xF0F0F0F0F0);
	* @endcode
	*
	* @param address The 40-bit address of the pipe to open.
	*/
	void openWritingPipe(uint64_t address);

	/**
	* Empty the receive buffer
	*
	* @return Current value of status register
	*/
	uint8_t flush_rx(void);

private:

	/**
	* @name Low-level internal interface.
	*
	*  Protected methods that address the chip directly.  Regular users cannot
	*  ever call these.  They are documented for completeness and for developers who
	*  may want to extend this class.
	*/
	/**@{*/

	/**
	* Set chip select pin
	*
	* Running SPI bus at PI_CLOCK_DIV2 so we don't waste time transferring data
	* and best of all, we make use of the radio's FIFO buffers. A lower speed
	* means we're less likely to effectively leverage our FIFOs and pay a higher
	* AVR runtime cost as toll.
	*
	* @param mode HIGH to take this unit off the SPI bus, LOW to put it on
	*/
	void csn(bool mode);

	/**
	* Set chip enable
	*
	* @param level HIGH to actively begin transmission or LOW to put in standby.  Please see data sheet
	* for a much more detailed description of this pin.
	*/
	void ce(bool level);

	/**
	* Read a chunk of data in from a register
	*
	* @param reg Which register. Use constants from nRF24L01.h
	* @param buf Where to put the data
	* @param len How many bytes of data to transfer
	* @return Current value of status register
	*/
	uint8_t read_register(uint8_t reg, uint8_t* buf, uint8_t len);

	/**
	* Read single byte from a register
	*
	* @param reg Which register. Use constants from nRF24L01.h
	* @return Current value of register @p reg
	*/
	uint8_t read_register(uint8_t reg);

	/**
	* Write a chunk of data to a register
	*
	* @param reg Which register. Use constants from nRF24L01.h
	* @param buf Where to get the data
	* @param len How many bytes of data to transfer
	* @return Current value of status register
	*/
	uint8_t write_register(uint8_t reg, const uint8_t* buf, uint8_t len);

	/**
	* Write a single byte to a register
	*
	* @param reg Which register. Use constants from nRF24L01.h
	* @param value The new value to write
	* @return Current value of status register
	*/
	uint8_t write_register(uint8_t reg, uint8_t value);

	/**
	* Write the transmit payload
	*
	* The size of data written is the fixed payload size, see getPayloadSize()
	*
	* @param buf Where to get the data
	* @param len Number of bytes to be sent
	* @return Current value of status register
	*/
	uint8_t write_payload(const void* buf, uint8_t len, const uint8_t writeType);

	/**
	* Read the receive payload
	*
	* The size of data read is the fixed payload size, see getPayloadSize()
	*
	* @param buf Where to put the data
	* @param len Maximum number of bytes to read
	* @return Current value of status register
	*/
	uint8_t read_payload(void* buf, uint8_t len);

	/**
	* Retrieve the current status of the chip
	*
	* @return Current value of status register
	*/
	uint8_t get_status(void);

#if !defined (MINIMAL)
	/**
	* Decode and print the given status to stdout
	*
	* @param status Status value to print
	*
	* @warning Does nothing if stdout is not defined.  See fdevopen in stdio.h
	*/
	void print_status(uint8_t status);

	/**
	* Decode and print the given 'observe_tx' value to stdout
	*
	* @param value The observe_tx value to print
	*
	* @warning Does nothing if stdout is not defined.  See fdevopen in stdio.h
	*/
	void print_observe_tx(uint8_t value);

	/**
	* Print the name and value of an 8-bit register to stdout
	*
	* Optionally it can print some quantity of successive
	* registers on the same line.  This is useful for printing a group
	* of related registers on one line.
	*
	* @param name Name of the register
	* @param reg Which register. Use constants from nRF24L01.h
	* @param qty How many successive registers to print
	*/
	void print_byte_register(const char* name, uint8_t reg, uint8_t qty = 1);

	/**
	* Print the name and value of a 40-bit address register to stdout
	*
	* Optionally it can print some quantity of successive
	* registers on the same line.  This is useful for printing a group
	* of related registers on one line.
	*
	* @param name Name of the register
	* @param reg Which register. Use constants from nRF24L01.h
	* @param qty How many successive registers to print
	*/
	void print_address_register(const char* name, uint8_t reg, uint8_t qty = 1);
#endif
	/**
	* Turn on or off the special features of the chip
	*
	* The chip has certain 'features' which are only available when the 'features'
	* are enabled.  See the datasheet for details.
	*/
	void toggle_features(void);

	/**
	* Built in spi transfer function to simplify repeating code repeating code
	*/

	uint8_t spiTrans(uint8_t cmd);

#if defined (FAILURE_HANDLING) || defined (RF24_LINUX)
	void errNotify(void);
#endif

#pragma endregion RF24 public

#pragma region RF24
	/**
	* SPI transactions
	*
	* Common code for SPI transactions including CSN toggle
	*
	*/
	inline void beginTransaction();

	inline void endTransaction();
#pragma endregion RF24 protected

#pragma region RF24
	uint32_t signalTimer;
	uint8_t userChannel;
	uint16_t serialNumber;
#ifdef SOFTSPI
	SoftSPI<SOFT_SPI_MISO_PIN, SOFT_SPI_MOSI_PIN, SOFT_SPI_SCK_PIN, SPI_MODE> spi;
#elif defined (SPI_UART)
	SPIUARTClass uspi;
#endif

#if defined (RF24_LINUX) || defined (XMEGA_D3) /* XMEGA can use SPI class */
	SPI spi;
#endif
#if defined (MRAA)
	GPIO gpio;
#endif

	uint16_t ce_pin; /**< "Chip Enable" pin, activates the RX or TX role */
	uint16_t csn_pin; /**< SPI Chip select */
	uint16_t spi_speed; /**< SPI Bus Speed */
#if defined (RF24_LINUX) || defined (XMEGA_D3)
	uint8_t spi_rxbuff[32 + 1]; //SPI receive buffer (payload max 32 bytes)
	uint8_t spi_txbuff[32 + 1]; //SPI transmit buffer (payload max 32 bytes + 1 byte for the command)
#endif  
	bool p_variant; /* False for RF24L01 and true for RF24L01P */
	uint8_t payload_size; /**< Fixed size of payloads */
	bool dynamic_payloads_enabled; /**< Whether dynamic payloads are enabled. */
	uint8_t pipe0_reading_address[5]; /**< Last address set on pipe 0 for reading. */
	uint8_t addr_width; /**< The address width to use - 3,4 or 5 bytes. */
	uint8_t payload[MAX_PAYLOAD_SIZE];
	void updateRF24();
	uint8_t ack;
	void handleRF24Packet();
	uint8_t joystickX;
	uint8_t joystickY;
	uint8_t previousButtonState;
	uint8_t buttonState;
	bool button(uint8_t btn);
	bool pressed(uint8_t btn);
	bool released(uint8_t btn);
#pragma endregion RF24 private

#pragma region Custom RF24
	void offline();
	void online();
	void reconnect();
	bool isOnline = false;
	uint8_t rf24ReceiveCounter = 0;
	uint32_t offlineTimer;
	bool handlePacket = true;
	bool rf24HasBegin = false;
	void delayWithRF24Update(uint16_t duration);
#pragma endregion Custom RF24

#pragma region Communication and Control
	void commandTank();
	uint8_t maxForwardSpeed;
	uint8_t maxBackwardSpeed;
	uint8_t maxTurnSpeed;
	uint8_t maxSpeed;
	void handleButtons();
	TANK_STATE tankState = TANK_STATE_NULL;
	int8_t reviveMotorCommandValue = -60;
#pragma endregion Communication and Control

#pragma region Game play
	uint16_t reloadingPeriod = DEFAULT_RELOADING_PERIOD;
	bool isReloading = false;
	uint32_t reloadingTimer;
	void updateGamePlay();
	int16_t blood = DEFAULT_BLOOD;
	BLOOD_STATE bloodState = BLOOD_STATE_NORMAL;
	int8_t life = DEFAULT_LIFE;
	bool flashHeadLed = false;
	uint32_t bloodStateFlashingTimer;
	uint32_t bloodStateFlashingDuration;
	bool headLedIsOn = true;
	bool isInvincible = false;
	uint32_t invincibleTimer;
	uint32_t invincibleDuration = DEFAULT_INVINCIBLE_PERIOD_AFTER_HIT;
	bool isGameOver = false;
	bool canBeHitAfterGameoverRegister = true;
	void fire();
	void dead();
	void revive();
	void gameOver();
	void hit();
	void hitAfterGameover();
	void selectGameMode();
	GAME_MODE gameMode;
	void toogleGameMode();
	uint8_t headColorIndex = 1;
	uint8_t eyesColorIndex = 1;
	void toogleHeadColor();
	void toogleEyesColor();
	void hitByTeammate();
	uint8_t myID;
#pragma endregion Game play

#pragma region IR

	struct IR_NEC_dataContainer
	{
		union IR_NEC_address
		{
			uint16_t address;
			struct
			{
				uint8_t byte1;
				uint8_t byte2;
			};

		}NEC_address;
		union IR_NEC_command
		{
			uint32_t command;
			struct
			{
				uint8_t commandNot;
				uint8_t byte3;
			};
		}NEC_command;
	};


	IR_NEC_dataContainer irSendPacket;
	IR_NEC_dataContainer irReceivePacket;
	void handleIRPacket();
	void updateIRReceiver();

#pragma endregion IR

};

#pragma endregion Tank

#pragma region Joystick



#pragma endregion Joystick

#pragma region Booster Pad
#pragma endregion Booster Pad



/*

#### v2.0.0
- [Major Change] Joystick library incorporated into this tank libary

*/



#pragma region Something
#pragma endregion Something


#endif //_Fuzzy_Tank_h

